import uuid
import os

from dagster import (
    ResourceDefinition,
    InputDefinition,
    Int,
    ModeDefinition,
    OutputDefinition,
    execute_pipeline,
    pipeline,
    solid,
)
from dagster.seven import mock

from dagster_aws.s3.system_storage import s3_plus_default_storage_defs
from dagster_aws.s3.file_manager import S3FileManager, S3FileHandle
from dagster_aws.s3.s3_fake_resource import create_s3_fake_resource

# For deps
# pylint: disable=no-value-for-parameter


def test_s3_file_manager_write():
    s3_mock = mock.MagicMock()
    file_manager = S3FileManager(s3_mock, 'some-bucket', 'some-key')

    foo_bytes = 'foo'.encode()

    file_handle = file_manager.write_data(foo_bytes)

    assert isinstance(file_handle, S3FileHandle)

    assert file_handle.s3_bucket == 'some-bucket'
    assert file_handle.s3_key.startswith('some-key/')

    assert s3_mock.put_object.call_count == 1


def test_s3_file_manager_read():
    state = {'called': 0}
    bar_bytes = 'bar'.encode()

    class S3Mock(mock.MagicMock):
        def download_file(self, *_args, **kwargs):
            state['called'] += 1
            assert state['called'] == 1
            state['bucket'] = kwargs.get('Bucket')
            state['key'] = kwargs.get('Key')
            file_name = kwargs.get('Filename')
            state['file_name'] = file_name
            with open(file_name, 'wb') as ff:
                ff.write(bar_bytes)

    s3_mock = S3Mock()
    file_manager = S3FileManager(s3_mock, 'some-bucket', 'some-key')
    file_handle = S3FileHandle('some-bucket', 'some-key/kdjfkjdkfjkd')
    with file_manager.read(file_handle) as file_obj:
        assert file_obj.read() == bar_bytes

    assert state['bucket'] == file_handle.s3_bucket
    assert state['key'] == file_handle.s3_key

    # read again. cached
    with file_manager.read(file_handle) as file_obj:
        assert file_obj.read() == bar_bytes

    assert os.path.exists(state['file_name'])

    file_manager.delete_local_temp()

    assert not os.path.exists(state['file_name'])


def test_depends_on_s3_resource_intermediates():
    @solid(
        input_defs=[InputDefinition('num_one', Int), InputDefinition('num_two', Int)],
        output_defs=[OutputDefinition(Int)],
    )
    def add_numbers(_, num_one, num_two):
        return num_one + num_two

    s3_fake_resource = create_s3_fake_resource()

    @pipeline(
        mode_defs=[
            ModeDefinition(
                system_storage_defs=s3_plus_default_storage_defs,
                resource_defs={'s3': ResourceDefinition.hardcoded_resource(s3_fake_resource)},
            )
        ]
    )
    def s3_internal_pipeline():
        return add_numbers()

    result = execute_pipeline(
        s3_internal_pipeline,
        environment_dict={
            'solids': {
                'add_numbers': {'inputs': {'num_one': {'value': 2}, 'num_two': {'value': 4}}}
            },
            'storage': {'s3': {'config': {'s3_bucket': 'some-bucket'}}},
        },
    )

    assert result.success
    assert result.result_for_solid('add_numbers').output_value() == 6

    assert 'some-bucket' in s3_fake_resource.session.buckets

    keys = set()
    for step_key, output_name in [('add_numbers.compute', 'result')]:
        keys.add(create_s3_key(result.run_id, step_key, output_name))

    assert set(s3_fake_resource.session.buckets['some-bucket'].keys()) == keys


def create_s3_key(run_id, step_key, output_name):
    return 'dagster/runs/{run_id}/files/intermediates/{step_key}/{output_name}'.format(
        run_id=run_id, step_key=step_key, output_name=output_name
    )


def test_depends_on_s3_resource_file_manager():
    bar_bytes = 'bar'.encode()

    @solid(output_defs=[OutputDefinition(S3FileHandle)])
    def emit_file(context):
        return context.file_manager.write_data(bar_bytes)

    @solid(input_defs=[InputDefinition('file_handle', S3FileHandle)])
    def accept_file(context, file_handle):
        local_path = context.file_manager.copy_handle_to_local_temp(file_handle)
        assert isinstance(local_path, str)
        assert open(local_path, 'rb').read() == bar_bytes

    s3_fake_resource = create_s3_fake_resource()

    @pipeline(
        mode_defs=[
            ModeDefinition(
                system_storage_defs=s3_plus_default_storage_defs,
                resource_defs={'s3': ResourceDefinition.hardcoded_resource(s3_fake_resource)},
            )
        ]
    )
    def s3_file_manager_test():
        accept_file(emit_file())

    result = execute_pipeline(
        s3_file_manager_test,
        environment_dict={'storage': {'s3': {'config': {'s3_bucket': 'some-bucket'}}}},
    )

    assert result.success

    keys_in_bucket = set(s3_fake_resource.session.buckets['some-bucket'].keys())

    for step_key, output_name in [
        ('emit_file.compute', 'result'),
        ('accept_file.compute', 'result'),
    ]:
        keys_in_bucket.remove(create_s3_key(result.run_id, step_key, output_name))

    assert len(keys_in_bucket) == 1

    file_key = list(keys_in_bucket)[0]
    comps = file_key.split('/')

    assert '/'.join(comps[:-1]) == 'dagster/runs/{run_id}/files/managed'.format(
        run_id=result.run_id
    )

    assert uuid.UUID(comps[-1])



from collections import defaultdict
import io

from dagster_aws.s3.resources import S3Resource
from dagster.seven import mock

from botocore.exceptions import ClientError


def create_s3_fake_resource():
    return S3Resource(S3FakeSession())


class S3FakeSession:
    def __init__(self, buckets=None):
        self.buckets = defaultdict(dict, buckets) if buckets else defaultdict(dict)
        self.mock_extras = mock.MagicMock()

    def head_bucket(self, Bucket, *args, **kwargs):  # pylint: disable=unused-argument
        self.mock_extras.head_bucket(*args, **kwargs)

    def head_object(self, Bucket, Key, *args, **kwargs):
        self.mock_extras.head_object(*args, **kwargs)
        return {'ContentLength': len(self.buckets.get(Bucket, {}).get(Key, b''))}

    def list_objects_v2(self, Bucket, Prefix, *args, **kwargs):
        self.mock_extras.list_objects_v2(*args, **kwargs)
        key = self.buckets.get(Bucket, {}).get(Prefix)
        if key:
            return {'KeyCount': 1, 'Contents': [{'Key': key}], 'IsTruncated': False}
        else:
            return {'KeyCount': 0, 'Contents': [], 'IsTruncated': False}

    def put_object(self, Bucket, Key, Body, *args, **kwargs):
        self.mock_extras.put_object(*args, **kwargs)
        self.buckets[Bucket][Key] = Body.read()

    def get_object(self, Bucket, Key, *args, **kwargs):
        if not self._has_object(Bucket, Key):
            raise ClientError({}, None)

        self.mock_extras.get_object(*args, **kwargs)
        return {'Body': self._get_byte_stream(Bucket, Key)}

    def _has_object(self, bucket, key):
        return bucket in self.buckets and key in self.buckets[bucket]

    def _get_byte_stream(self, bucket, key):
        return io.BytesIO(self.buckets[bucket][key])

    def download_file(self, Bucket, Key, Filename, *args, **kwargs):
        self.mock_extras.download_file(*args, **kwargs)
        with open(Filename, 'wb') as ff:
            ff.write(self._get_byte_stream(Bucket, Key).read())



import datetime
import json

import boto3
import pytest

from dagster import execute_pipeline, ModeDefinition, PipelineDefinition, solid
from dagster_aws.cloudwatch import cloudwatch_logger
from dagster_aws.cloudwatch.loggers import millisecond_timestamp


TEST_CLOUDWATCH_LOG_GROUP_NAME = '/dagster-test/test-cloudwatch-logging'
TEST_CLOUDWATCH_LOG_STREAM_NAME = 'test-logging'


def permissioned_and_connected():
    try:
        client = boto3.client('logs', 'us-west-1')
        client.describe_log_groups(logGroupNamePrefix=TEST_CLOUDWATCH_LOG_GROUP_NAME)
        res = client.describe_log_streams(
            logGroupName=TEST_CLOUDWATCH_LOG_GROUP_NAME,
            logStreamNamePrefix=TEST_CLOUDWATCH_LOG_STREAM_NAME,
        )
        if TEST_CLOUDWATCH_LOG_STREAM_NAME in (
            log_stream['logStreamName'] for log_stream in res['logStreams']
        ):
            return True
        else:
            return False
    except:  # pylint: disable=bare-except
        return False


cloudwatch_test = pytest.mark.skipif(
    not permissioned_and_connected(),
    reason='Not credentialed for cloudwatch tests or could not connect to AWS',
)


@solid(name='hello_cloudwatch')
def hello_cloudwatch(context):
    context.log.info('Hello, Cloudwatch!')


def define_hello_cloudwatch_pipeline():
    return PipelineDefinition(
        [hello_cloudwatch],
        name='hello_cloudwatch_pipeline',
        mode_defs=[ModeDefinition(logger_defs={'cloudwatch': cloudwatch_logger})],
    )


@cloudwatch_test
def test_cloudwatch_logging_bad_log_group_name():
    with pytest.raises(
        Exception,
        match='Failed to initialize Cloudwatch logger: Could not find log group with name foo',
    ):
        execute_pipeline(
            define_hello_cloudwatch_pipeline(),
            {
                'loggers': {
                    'cloudwatch': {
                        'config': {
                            'log_group_name': 'foo',
                            'log_stream_name': 'bar',
                            'aws_region': 'us-east-1',
                        }
                    }
                }
            },
        )


@cloudwatch_test
def test_cloudwatch_logging_bad_log_stream_name():
    with pytest.raises(
        Exception,
        match='Failed to initialize Cloudwatch logger: Could not find log stream with name bar',
    ):
        execute_pipeline(
            define_hello_cloudwatch_pipeline(),
            {
                'loggers': {
                    'cloudwatch': {
                        'config': {
                            'log_group_name': TEST_CLOUDWATCH_LOG_GROUP_NAME,
                            'log_stream_name': 'bar',
                            'aws_region': 'us-west-1',
                        }
                    }
                }
            },
        )


# TODO: Test bad region


@cloudwatch_test
def test_cloudwatch_logging():
    res = execute_pipeline(
        define_hello_cloudwatch_pipeline(),
        {
            'loggers': {
                'cloudwatch': {
                    'config': {
                        'log_group_name': TEST_CLOUDWATCH_LOG_GROUP_NAME,
                        'log_stream_name': TEST_CLOUDWATCH_LOG_STREAM_NAME,
                        'aws_region': 'us-west-1',
                    }
                }
            }
        },
    )

    client = boto3.client('logs', 'us-west-1')

    now = millisecond_timestamp(datetime.datetime.utcnow())

    # This is implicitly assuming that we're not running these tests with too much concurrency, etc.
    events = client.get_log_events(
        endTime=now,
        logGroupName=TEST_CLOUDWATCH_LOG_GROUP_NAME,
        logStreamName=TEST_CLOUDWATCH_LOG_STREAM_NAME,
        limit=100,
    )['events']

    found_orig_message = False
    for parsed_msg in (json.loads(event['message']) for event in events):
        if parsed_msg['dagster_meta']['run_id'] == res.run_id:
            if parsed_msg['dagster_meta']['orig_message'] == 'Hello, Cloudwatch!':
                found_orig_message = True
                break

    assert found_orig_message



from dagster import check
from dagster.core.storage.intermediate_store import IntermediateStore
from dagster.core.storage.type_storage import TypeStoragePluginRegistry
from .object_store import S3ObjectStore


class S3IntermediateStore(IntermediateStore):
    def __init__(self, s3_bucket, run_id, s3_session=None, type_storage_plugin_registry=None):
        check.str_param(s3_bucket, 's3_bucket')
        check.str_param(run_id, 'run_id')

        object_store = S3ObjectStore(s3_bucket, s3_session=s3_session)
        root = object_store.key_for_paths(['dagster', 'runs', run_id, 'files'])

        super(S3IntermediateStore, self).__init__(
            object_store,
            root,
            type_storage_plugin_registry=check.inst_param(
                type_storage_plugin_registry
                if type_storage_plugin_registry
                else TypeStoragePluginRegistry(types_to_register={}),
                'type_storage_plugin_registry',
                TypeStoragePluginRegistry,
            ),
        )

    def copy_object_from_prev_run(
        self, context, previous_run_id, paths
    ):  # pylint: disable=unused-argument
        check.not_implemented('not supported: TODO for max. put issue number here')



from enum import Enum as PyEnum

from dagster import Enum, EnumValue

EbsVolumeType = Enum(
    name='EbsVolumeType', enum_values=[EnumValue('gp2'), EnumValue('io1'), EnumValue('standard')]
)


class EmrClusterState(PyEnum):
    Starting = 'STARTING'
    Bootstrapping = 'BOOTSTRAPPING'
    Running = 'RUNNING'
    Waiting = 'WAITING'
    Terminating = 'TERMINATING'
    Terminated = 'TERMINATED'
    TerminatedWithErrors = 'TERMINATED_WITH_ERRORS'


EmrActionOnFailure = Enum(
    name='EmrActionOnFailure',
    enum_values=[
        EnumValue('TERMINATE_JOB_FLOW'),
        EnumValue('TERMINATE_CLUSTER'),
        EnumValue('CANCEL_AND_WAIT'),
        EnumValue('CONTINUE'),
    ],
)

EmrAdjustmentType = Enum(
    name='EmrAdjustmentType',
    enum_values=[
        EnumValue('CHANGE_IN_CAPACITY'),
        EnumValue('PERCENT_CHANGE_IN_CAPACITY'),
        EnumValue('EXACT_CAPACITY'),
    ],
)

EmrComparisonOperator = Enum(
    name='EmrComparisonOperator',
    enum_values=[
        EnumValue('GREATER_THAN_OR_EQUAL'),
        EnumValue('GREATER_THAN'),
        EnumValue('LESS_THAN'),
        EnumValue('LESS_THAN_OR_EQUAL'),
    ],
)

EmrInstanceRole = Enum(
    name='EmrInstanceRole', enum_values=[EnumValue('MASTER'), EnumValue('CORE'), EnumValue('TASK')]
)

EmrMarket = Enum(name='EmrMarket', enum_values=[EnumValue('ON_DEMAND'), EnumValue('SPOT')])

EmrRepoUpgradeOnBoot = Enum(
    name='EmrRepoUpgradeOnBoot', enum_values=[EnumValue('SECURITY'), EnumValue('NONE')]
)

EmrScaleDownBehavior = Enum(
    name='EmrScaleDownBehavior',
    enum_values=[
        EnumValue('TERMINATE_AT_INSTANCE_HOUR'),
        EnumValue('TERMINATE_AT_TASK_COMPLETION'),
    ],
)

EmrStatistic = Enum(
    name='EmrStatistic',
    enum_values=[
        EnumValue('SAMPLE_COUNT'),
        EnumValue('AVERAGE'),
        EnumValue('SUM'),
        EnumValue('MINIMUM'),
        EnumValue('MAXIMUM'),
    ],
)

EmrSupportedProducts = Enum(
    name='EmrSupportedProducts', enum_values=[EnumValue('mapr-m3'), EnumValue('mapr-m5')]
)

EmrTimeoutAction = Enum(
    name='EmrTimeoutAction',
    enum_values=[EnumValue('SWITCH_TO_ON_DEMAND'), EnumValue('TERMINATE_CLUSTER')],
)

EmrUnit = Enum(
    name='EmrUnit',
    enum_values=[
        EnumValue('NONE'),
        EnumValue('SECONDS'),
        EnumValue('MICRO_SECONDS'),
        EnumValue('MILLI_SECONDS'),
        EnumValue('BYTES'),
        EnumValue('KILO_BYTES'),
        EnumValue('MEGA_BYTES'),
        EnumValue('GIGA_BYTES'),
        EnumValue('TERA_BYTES'),
        EnumValue('BITS'),
        EnumValue('KILO_BITS'),
        EnumValue('MEGA_BITS'),
        EnumValue('GIGA_BITS'),
        EnumValue('TERA_BITS'),
        EnumValue('PERCENT'),
        EnumValue('COUNT'),
        EnumValue('BYTES_PER_SECOND'),
        EnumValue('KILO_BYTES_PER_SECOND'),
        EnumValue('MEGA_BYTES_PER_SECOND'),
        EnumValue('GIGA_BYTES_PER_SECOND'),
        EnumValue('TERA_BYTES_PER_SECOND'),
        EnumValue('BITS_PER_SECOND'),
        EnumValue('KILO_BITS_PER_SECOND'),
        EnumValue('MEGA_BITS_PER_SECOND'),
        EnumValue('GIGA_BITS_PER_SECOND'),
        EnumValue('TERA_BITS_PER_SECOND'),
        EnumValue('COUNT_PER_SECOND'),
    ],
)



import logging

from io import BytesIO

import boto3

from dagster import check
from dagster.core.storage.object_store import ObjectStore
from dagster.core.types.marshal import SerializationStrategy


class S3ObjectStore(ObjectStore):
    def __init__(self, bucket, s3_session=None):
        self.bucket = check.str_param(bucket, 'bucket')
        self.s3 = s3_session or boto3.client('s3')
        self.s3.head_bucket(Bucket=bucket)
        super(S3ObjectStore, self).__init__(sep='/')

    def set_object(self, key, obj, serialization_strategy=None):
        check.str_param(key, 'key')
        # cannot check obj since could be arbitrary Python object
        check.inst_param(
            serialization_strategy, 'serialization_strategy', SerializationStrategy
        )  # cannot be none here

        if self.has_object(key):
            logging.warning('Removing existing S3 key: {key}'.format(key=key))
            self.rm_object(key)

        with BytesIO() as bytes_io:
            serialization_strategy.serialize(obj, bytes_io)
            bytes_io.seek(0)
            self.s3.put_object(Bucket=self.bucket, Key=key, Body=bytes_io)

        return self.uri_for_key(key)

    def get_object(self, key, serialization_strategy=None):
        check.str_param(key, 'key')
        check.param_invariant(len(key) > 0, 'key')

        # FIXME we need better error handling for object store
        return serialization_strategy.deserialize(
            BytesIO(self.s3.get_object(Bucket=self.bucket, Key=key)['Body'].read())
        )

    def has_object(self, key):
        check.str_param(key, 'key')
        check.param_invariant(len(key) > 0, 'key')

        key_count = self.s3.list_objects_v2(Bucket=self.bucket, Prefix=key)['KeyCount']
        return bool(key_count > 0)

    def rm_object(self, key):
        check.str_param(key, 'key')
        check.param_invariant(len(key) > 0, 'key')

        def delete_for_results(store, results):
            store.s3.delete_objects(
                Bucket=store.bucket,
                Delete={'Objects': [{'Key': result['Key']} for result in results['Contents']]},
            )

        if not self.has_object(key):
            return

        results = self.s3.list_objects_v2(Bucket=self.bucket, Prefix=key)
        delete_for_results(self, results)

        continuation = results['IsTruncated']
        while continuation:
            continuation_token = results['NextContinuationToken']
            results = self.s3.list_objects_v2(
                Bucket=self.bucket, Prefix=key, ContinuationToken=continuation_token
            )
            delete_for_results(self, results)
            continuation = results['IsTruncated']

        return

    def cp_object(self, src, dst):
        # https://github.com/dagster-io/dagster/issues/1455
        raise NotImplementedError()

    def uri_for_key(self, key, protocol=None):
        check.str_param(key, 'key')
        protocol = check.opt_str_param(protocol, 'protocol', default='s3://')
        return protocol + self.bucket + '/' + '{key}'.format(key=key)



import time

import boto3

from dagster import (
    check,
    Dict,
    Field,
    InputDefinition,
    Nothing,
    OutputDefinition,
    Output,
    SolidDefinition,
    String,
)

from .configs import define_emr_run_job_flow_config
from .types import EmrClusterState


_START = 'start'

# wait at most 24 hours by default for cluster job completion
_DEFAULT_CLUSTER_MAX_WAIT_TIME_SEC = 24 * 60 * 60


class EmrRunJobFlowSolidDefinition(SolidDefinition):
    def __init__(
        self,
        name,
        description=None,
        max_wait_time_sec=_DEFAULT_CLUSTER_MAX_WAIT_TIME_SEC,
        poll_interval_sec=5,
    ):
        name = check.str_param(name, 'name')

        description = check.opt_str_param(description, 'description', 'EMR create job flow solid.')

        def _compute_fn(context, _):  # pylint: disable=too-many-locals
            client = boto3.client('emr', region_name=context.solid_config.get('aws_region'))

            # kick off the EMR job flow
            response = client.run_job_flow(**context.solid_config['job_config'])

            # Job flow IDs and cluster IDs are interchangable
            job_flow_id = response.get('JobFlowId')

            context.log.info('waiting for EMR cluster job flow completion...')

            max_iter = max_wait_time_sec / poll_interval_sec

            # wait for the task
            done = False
            curr_iter = 0
            while not done and curr_iter < max_iter:
                cluster = client.describe_cluster(ClusterId=job_flow_id)
                status = cluster.get('Cluster', {}).get('Status', {})
                state = status.get('State')
                state_change_reason = status.get('StateChangeReason', {}).get('Message')

                context.log.info(
                    'EMR cluster %s state: %s state change reason: %s'
                    % (job_flow_id, state, state_change_reason)
                )

                # This will take a while... cluster creation usually > 5 minutes
                time.sleep(poll_interval_sec)

                # Note that the user can specify Instances.KeepJobFlowAliveWhenNoSteps, which will
                # keep the cluster alive after the job completes. In such cases where the cluster
                # continues in waiting state, we stop waiting here and yield.

                # See: https://bit.ly/2UUq1G9
                # pylint: disable=no-member
                done = state in {
                    EmrClusterState.Waiting.value,
                    EmrClusterState.Terminated.value,
                    EmrClusterState.TerminatedWithErrors.value,
                }

                curr_iter += 1

            yield Output(job_flow_id)

        super(EmrRunJobFlowSolidDefinition, self).__init__(
            name=name,
            description=description,
            input_defs=[InputDefinition(_START, Nothing)],
            output_defs=[OutputDefinition(String)],
            compute_fn=_compute_fn,
            config_field=Field(
                Dict(
                    {
                        'aws_region': Field(String, is_optional=True),
                        'job_config': define_emr_run_job_flow_config(),
                    }
                )
            ),
        )



from dagster import system_storage, SystemStorageData, Field, String
from dagster.core.definitions.system_storage import mem_system_storage, fs_system_storage
from dagster.core.storage.intermediates_manager import IntermediateStoreIntermediatesManager
from dagster.core.storage.runs import FileSystemRunStorage
from .intermediate_store import S3IntermediateStore
from .file_manager import S3FileManager


@system_storage(
    name='s3',
    is_persistent=True,
    config={'s3_bucket': Field(String)},
    required_resource_keys={'s3'},
)
def s3_system_storage(init_context):
    s3_session = init_context.resources.s3.session
    s3_key = 'dagster/runs/{run_id}/files/managed'.format(run_id=init_context.run_config.run_id)
    return SystemStorageData(
        file_manager=S3FileManager(
            s3_session=s3_session,
            s3_bucket=init_context.system_storage_config['s3_bucket'],
            s3_base_key=s3_key,
        ),
        run_storage=FileSystemRunStorage(),
        intermediates_manager=IntermediateStoreIntermediatesManager(
            S3IntermediateStore(
                s3_session=s3_session,
                s3_bucket=init_context.system_storage_config['s3_bucket'],
                run_id=init_context.run_config.run_id,
                type_storage_plugin_registry=init_context.type_storage_plugin_registry,
            )
        ),
    )


s3_plus_default_storage_defs = [mem_system_storage, fs_system_storage, s3_system_storage]



from .version import __version__
from .s3.solids import S3Coordinate
from .s3.resources import s3_resource, S3Resource
from .s3.s3_fake_resource import S3FakeSession
from .s3.file_cache import S3FileCache



import datetime
import logging

import boto3

from dagster import check, Field, logger, seven
from dagster.core.log_manager import coerce_valid_log_level

# The maximum batch size is 1,048,576 bytes, and this size is calculated as the sum of all event
# messages in UTF-8, plus 26 bytes for each log event.
MAXIMUM_BATCH_SIZE = 1048576
OVERHEAD = 26

EPOCH = datetime.datetime(1970, 1, 1)

# For real
def millisecond_timestamp(dt):
    td = dt - EPOCH
    microsecond_timestamp = (
        td.days * 24 * 60 * 60 * 1000000 + td.seconds * 1000000 + td.microseconds
    )
    return int(microsecond_timestamp / 1000)


class CloudwatchLogsHandler(logging.Handler):
    def __init__(
        self,
        log_group_name,
        log_stream_name,
        aws_region=None,
        aws_secret_access_key=None,
        aws_access_key_id=None,
    ):
        self.client = boto3.client(
            'logs',
            region_name=aws_region,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
        )
        self.log_group_name = check.str_param(log_group_name, 'log_group_name')
        # Maybe we should make this optional, and default to the run_id
        self.log_stream_name = check.str_param(log_stream_name, 'log_stream_name')
        self.overhead = OVERHEAD
        self.maximum_batch_size = MAXIMUM_BATCH_SIZE
        self.sequence_token = None

        self.check_log_group()
        self.check_log_stream()

        super(CloudwatchLogsHandler, self).__init__()

    def check_log_group(self):
        # Check that log group exists
        log_group_exists = False
        next_token = None
        while not log_group_exists:
            describe_log_group_kwargs = {'logGroupNamePrefix': self.log_group_name}
            if next_token is not None:
                describe_log_group_kwargs['nextToken'] = next_token

            res = self.client.describe_log_groups(**describe_log_group_kwargs)
            if self.log_group_name in (log_group['logGroupName'] for log_group in res['logGroups']):
                log_group_exists = True
                break
            else:
                next_token = res.get('nextToken')
                if next_token is None:
                    break

        if not log_group_exists:
            raise Exception(
                'Failed to initialize Cloudwatch logger: Could not find log group with name '
                '{log_group_name}'.format(log_group_name=self.log_group_name)
            )

    def check_log_stream(self):
        # Check that log stream exists
        log_stream_exists = False
        next_token = None
        while not log_stream_exists:
            describe_log_stream_kwargs = {
                'logGroupName': self.log_group_name,
                'logStreamNamePrefix': self.log_stream_name,
            }
            if next_token is not None:
                describe_log_stream_kwargs['nextToken'] = next_token

            res = self.client.describe_log_streams(**describe_log_stream_kwargs)
            for log_stream in res['logStreams']:
                if self.log_stream_name == log_stream['logStreamName']:
                    log_stream_exists = True
                    self.sequence_token = log_stream.get('uploadSequenceToken')
                break
            else:
                next_token = res.get('nextToken')
                if next_token is None:
                    break

        if not log_stream_exists:
            raise Exception(
                'Failed to initialize Cloudwatch logger: Could not find log stream with name '
                '{log_stream_name}'.format(log_stream_name=self.log_stream_name)
            )

    def log_error(self, record, exc):
        logging.critical('Error while logging!')
        try:
            logging.error(
                'Attempted to log: {record}'.format(record=seven.json.dumps(record.__dict__))
            )
        except Exception:  # pylint: disable=broad-except
            pass
        logging.exception(str(exc))

    def emit(self, record):
        self._emit(record, retry=False)

    def retry(self, record):
        self._emit(record, retry=True)

    def _emit(self, record, retry=False):
        message = seven.json.dumps(record.__dict__)
        timestamp = millisecond_timestamp(
            datetime.datetime.strptime(record.dagster_meta['log_timestamp'], '%Y-%m-%dT%H:%M:%S.%f')
        )
        params = {
            'logGroupName': self.log_group_name,
            'logStreamName': self.log_stream_name,
            'logEvents': [{'timestamp': timestamp, 'message': message}],
        }
        if self.sequence_token is not None:
            params['sequenceToken'] = self.sequence_token

        try:
            res = self.client.put_log_events(**params)
            self.sequence_token = res['nextSequenceToken']
            log_events_rejected = res.get('rejectedLogEventsInfo')
            if log_events_rejected is not None:
                logging.error('Cloudwatch logger: log events rejected: {res}'.format(res=res))
        except self.client.exceptions.InvalidSequenceTokenException as exc:
            if not retry:
                self.check_log_stream()
                self.retry(record)
            else:
                self.log_error(record, exc)
        except self.client.exceptions.DataAlreadyAcceptedException as exc:
            logging.error('Cloudwatch logger: log events already accepted: {res}'.format(res=res))
        except self.client.exceptions.InvalidParameterException as exc:
            logging.error(
                'Cloudwatch logger: Invalid parameter exception while logging: {res}'.format(
                    res=res
                )
            )
        except self.client.exceptions.ResourceNotFoundException as exc:
            logging.error(
                'Cloudwatch logger: Resource not found. Check that the log stream or log group '
                'was not deleted: {res}'.format(res=res)
            )
        except self.client.exceptions.ServiceUnavailableException as exc:
            if not retry:
                self.retry(record)
            else:
                logging.error('Cloudwatch logger: Service unavailable: {res}'.format(res=res))
        except self.client.exceptions.ServiceUnavailableException as exc:
            if not retry:
                self.retry(record)
            else:
                logging.error(
                    'Cloudwatch logger: Unrecognized client. Check your AWS access key id and '
                    'secret key: {res}'.format(res=res)
                )


@logger(
    {
        'log_level': Field(str, is_optional=True, default_value='INFO'),
        'name': Field(str, is_optional=True, default_value='dagster'),
        'log_group_name': Field(str),
        'log_stream_name': Field(str),
        'aws_region': Field(str, is_optional=True),
        'aws_secret_access_key': Field(str, is_optional=True, is_secret=True),
        'aws_access_key_id': Field(str, is_optional=True, is_secret=True),
    },
    description='The default colored console logger.',
)
def cloudwatch_logger(init_context):

    level = coerce_valid_log_level(init_context.logger_config['log_level'])
    name = init_context.logger_config['name']

    klass = logging.getLoggerClass()
    logger_ = klass(name, level=level)

    logger_.addHandler(
        CloudwatchLogsHandler(
            init_context.logger_config['log_group_name'],
            init_context.logger_config['log_stream_name'],
            aws_region=init_context.logger_config.get('aws_region'),
            aws_secret_access_key=init_context.logger_config.get('aws_secret_access_key'),
            aws_access_key_id=init_context.logger_config.get('aws_access_key_id'),
        )
    )
    return logger_



from dagster import Dict, Field, Int, PermissiveDict, String

from .types import S3ACL


def put_object_configs():
    return Field(
        Dict(
            fields={
                'ACL': Field(
                    S3ACL, description='The canned ACL to apply to the object.', is_optional=True
                ),
                # Body will be set by the solid, not supplied in config
                'Bucket': Field(
                    String,
                    description='Name of the bucket to which the PUT operation was initiated.',
                    is_optional=False,
                ),
                'CacheControl': Field(
                    String,
                    description='Specifies caching behavior along the request/reply chain.',
                    is_optional=True,
                ),
                'ContentDisposition': Field(
                    String,
                    description='Specifies presentational information for the object.',
                    is_optional=True,
                ),
                'ContentEncoding': Field(
                    String,
                    description='''Specifies what content encodings have been applied to the object
                    and thus what decoding mechanisms must be applied to obtain the media-type
                    referenced by the Content-Type header field.''',
                    is_optional=True,
                ),
                'ContentLanguage': Field(
                    String, description='The language the content is in.', is_optional=True
                ),
                'ContentLength': Field(
                    Int,
                    description='''Size of the body in bytes. This parameter is useful when the size
                    of the body cannot be determined automatically.''',
                    is_optional=True,
                ),
                'ContentMD5': Field(
                    String,
                    description='''The base64-encoded 128-bit MD5 digest of the part data. This
                    parameter is auto-populated when using the command from the CLI''',
                    is_optional=True,
                ),
                'ContentType': Field(
                    String,
                    description='A standard MIME type describing the format of the object data.',
                    is_optional=True,
                ),
                # TODO: datetime object
                # # 'Expires': Field(datetime, description='The date and time at which the object is
                # no longer cacheable.', is_optional=True),
                'GrantFullControl': Field(
                    String,
                    description='''Gives the grantee READ, READ_ACP, and WRITE_ACP permissions on
                    the object.''',
                    is_optional=True,
                ),
                'GrantRead': Field(
                    String,
                    description='Allows grantee to read the object data and its metadata.',
                    is_optional=True,
                ),
                'GrantReadACP': Field(
                    String, description='Allows grantee to read the object ACL.', is_optional=True
                ),
                'GrantWriteACP': Field(
                    String,
                    description='Allows grantee to write the ACL for the applicable object.',
                    is_optional=True,
                ),
                'Key': Field(
                    String,
                    description='Object key for which the PUT operation was initiated.',
                    is_optional=False,
                ),
                'Metadata': Field(
                    PermissiveDict(),
                    description='A map of metadata to store with the object in S3.',
                    is_optional=True,
                ),
                'ServerSideEncryption': Field(
                    String,
                    description='''The Server-side encryption algorithm used when storing this
                    object in S3 (e.g., AES256, aws:kms).''',
                    is_optional=True,
                ),
                'StorageClass': Field(
                    String,
                    description='''The type of storage to use for the object. Defaults to
                    'STANDARD'.''',
                    is_optional=True,
                ),
                'WebsiteRedirectLocation': Field(
                    String,
                    description='''If the bucket is configured as a website, redirects requests for
                    this object to another object in the same bucket or to an external URL. Amazon
                    S3 stores the value of this header in the object metadata.''',
                    is_optional=True,
                ),
                'SSECustomerAlgorithm': Field(
                    String,
                    description='''Specifies the algorithm to use to when encrypting the object
                    (e.g., AES256).''',
                    is_optional=True,
                ),
                'SSECustomerKey': Field(
                    String,
                    description='''Specifies the customer-provided encryption key for Amazon S3 to
                    use in encrypting data. This value is used to store the object and then it is
                    discarded; Amazon does not store the encryption key. The key must be appropriate
                    for use with the algorithm specified in the
                    x-amz-server-side-encryption-customer-algorithm header.''',
                    is_optional=True,
                ),
                'SSECustomerKeyMD5': Field(
                    String,
                    description='''Specifies the 128-bit MD5 digest of the encryption key according
                    to RFC 1321. Amazon S3 uses this header for a message integrity check to ensure
                    the encryption key was transmitted without error.

                    Please note that this parameter is automatically populated if it is not
                    provided. Including this parameter is not required''',
                    is_optional=True,
                ),
                'SSEKMSKeyId': Field(
                    String,
                    description='''Specifies the AWS KMS key ID to use for object encryption. All
                    GET and PUT requests for an object protected by AWS KMS will fail if not made
                    via SSL or using SigV4. Documentation on configuring any of the officially
                    supported AWS SDKs and CLI can be found at
                    http://docs.aws.amazon.com/AmazonS3/latest/dev/UsingAWSSDK.html#specify-signature-version''',
                    is_optional=True,
                ),
                'RequestPayer': Field(
                    String,
                    description='''Confirms that the requester knows that she or he will be charged
                    for the request. Bucket owners need not specify this parameter in their
                    requests. Documentation on downloading objects from requester pays buckets can
                    be found at
                    http://docs.aws.amazon.com/AmazonS3/latest/dev/ObjectsinRequesterPaysBuckets.html''',
                    is_optional=True,
                ),
                'Tagging': Field(
                    String,
                    description='''The tag-set for the object. The tag-set must be encoded as URL
                    Query parameters. (For example, "Key1=Value1")''',
                    is_optional=True,
                ),
                'ObjectLockMode': Field(
                    String,
                    description='The Object Lock mode that you want to apply to this object.',
                    is_optional=True,
                ),
                # TODO: datetime object 'ObjectLockRetainUntilDate': Field(datetime,
                # description='The date and time when you want this object\'s Object Lock to
                # expire.', is_optional=True),
                'ObjectLockLegalHoldStatus': Field(
                    String,
                    description='''The Legal Hold status that you want to apply to the specified
                    object.''',
                    is_optional=True,
                ),
            }
        )
    )



import pytest

from moto import mock_emr

from dagster import execute_pipeline, PipelineDefinition
from dagster_aws.emr.solids import EmrRunJobFlowSolidDefinition


@pytest.mark.skip
@mock_emr
def test_run_emr_job():
    e = EmrRunJobFlowSolidDefinition('test')
    pipeline = PipelineDefinition(name='test', solid_defs=[e])
    emr_config = {
        'Name': 'test-pyspark',
        'ReleaseLabel': 'emr-5.23.0',
        'Instances': {
            'MasterInstanceType': 'm4.large',
            'SlaveInstanceType': 'm4.large',
            'InstanceCount': 4,
            'TerminationProtected': False,
        },
        'Applications': [{'Name': 'Spark'}],
        'BootstrapActions': [
            {
                'Name': 'Spark Default Config',
                'ScriptBootstrapAction': {
                    'Path': 's3://support.elasticmapreduce/spark/maximize-spark-default-config'
                },
            }
        ],
        'Steps': [
            {
                'Name': 'Setup Debugging',
                'ActionOnFailure': 'TERMINATE_CLUSTER',
                'HadoopJarStep': {'Jar': 'command-runner.jar', 'Args': ['state-pusher-script']},
            },
            {
                'Name': 'setup - copy files',
                'ActionOnFailure': 'TERMINATE_CLUSTER',
                'HadoopJarStep': {
                    'Jar': 'command-runner.jar',
                    'Args': [
                        'aws',
                        's3',
                        'cp',
                        's3://elementl-public/pyspark/hello_world.py',
                        '/home/hadoop/',
                    ],
                },
            },
            {
                'Name': 'Run Spark',
                'ActionOnFailure': 'TERMINATE_CLUSTER',
                'HadoopJarStep': {
                    'Jar': 'command-runner.jar',
                    'Args': ['spark-submit', '/home/hadoop/main.py'],
                },
            },
        ],
        'VisibleToAllUsers': True,
        'JobFlowRole': 'EMR_EC2_DefaultRole',
        'ServiceRole': 'EMR_DefaultRole',
    }
    config = {'solids': {'test': {'config': {'job_config': emr_config, 'aws_region': 'us-east-1'}}}}
    result = execute_pipeline(pipeline, config)
    assert result.success



import os
import uuid

import pytest

from dagster import (
    Bool,
    DependencyDefinition,
    InputDefinition,
    Int,
    List,
    OutputDefinition,
    PipelineDefinition,
    RunConfig,
    SerializationStrategy,
    String,
    check,
    lambda_solid,
)

from dagster.core.events import DagsterEventType

from dagster.core.execution.api import create_execution_plan, execute_plan, scoped_pipeline_context

from dagster.core.storage.type_storage import TypeStoragePlugin

from dagster.core.types.runtime import (
    String as RuntimeString,
    resolve_to_runtime_type,
    RuntimeType,
    Bool as RuntimeBool,
)

from dagster.utils.test import yield_empty_pipeline_context

from dagster_aws.s3.intermediate_store import S3IntermediateStore


class UppercaseSerializationStrategy(SerializationStrategy):  # pylint: disable=no-init
    def serialize(self, value, write_file_obj):
        return write_file_obj.write(bytes(value.upper().encode('utf-8')))

    def deserialize(self, read_file_obj):
        return read_file_obj.read().decode('utf-8').lower()


class LowercaseString(RuntimeType):
    def __init__(self):
        super(LowercaseString, self).__init__(
            'lowercase_string',
            'LowercaseString',
            serialization_strategy=UppercaseSerializationStrategy(),
        )


def aws_credentials_present():
    return os.getenv('AWS_ACCESS_KEY_ID') and os.getenv('AWS_SECRET_ACCESS_KEY')


aws = pytest.mark.skipif(not aws_credentials_present(), reason='Couldn\'t find AWS credentials')

nettest = pytest.mark.nettest


def define_inty_pipeline():
    @lambda_solid
    def return_one():
        return 1

    @lambda_solid(input_defs=[InputDefinition('num', Int)], output_def=OutputDefinition(Int))
    def add_one(num):
        return num + 1

    @lambda_solid
    def user_throw_exception():
        raise Exception('whoops')

    pipeline = PipelineDefinition(
        name='basic_external_plan_execution',
        solid_defs=[return_one, add_one, user_throw_exception],
        dependencies={'add_one': {'num': DependencyDefinition('return_one')}},
    )
    return pipeline


def get_step_output(step_events, step_key, output_name='result'):
    for step_event in step_events:
        if (
            step_event.event_type == DagsterEventType.STEP_OUTPUT
            and step_event.step_key == step_key
            and step_event.step_output_data.output_name == output_name
        ):
            return step_event
    return None


@aws
@nettest
def test_using_s3_for_subplan(s3_bucket):
    pipeline = define_inty_pipeline()

    environment_dict = {'storage': {'s3': {'s3_bucket': s3_bucket}}}

    execution_plan = create_execution_plan(pipeline, environment_dict=environment_dict)

    assert execution_plan.get_step_by_key('return_one.compute')

    step_keys = ['return_one.compute']

    run_id = str(uuid.uuid4())

    store = S3IntermediateStore(s3_bucket, run_id)

    try:
        return_one_step_events = list(
            execute_plan(
                execution_plan,
                environment_dict=environment_dict,
                run_config=RunConfig(run_id=run_id),
                step_keys_to_execute=step_keys,
            )
        )

        assert get_step_output(return_one_step_events, 'return_one.compute')
        with scoped_pipeline_context(
            pipeline, environment_dict, RunConfig(run_id=run_id)
        ) as context:
            assert store.has_intermediate(context, 'return_one.compute')
            assert store.get_intermediate(context, 'return_one.compute', Int) == 1

        add_one_step_events = list(
            execute_plan(
                execution_plan,
                environment_dict=environment_dict,
                run_config=RunConfig(run_id=run_id),
                step_keys_to_execute=['add_one.compute'],
            )
        )

        assert get_step_output(add_one_step_events, 'add_one.compute')
        with scoped_pipeline_context(
            pipeline, environment_dict, RunConfig(run_id=run_id)
        ) as context:
            assert store.has_intermediate(context, 'add_one.compute')
            assert store.get_intermediate(context, 'add_one.compute', Int) == 2
    finally:
        with scoped_pipeline_context(
            pipeline, environment_dict, RunConfig(run_id=run_id)
        ) as context:
            store.rm_intermediate(context, 'return_one.compute')
            store.rm_intermediate(context, 'add_one.compute')


class FancyStringS3TypeStoragePlugin(TypeStoragePlugin):  # pylint:disable=no-init
    @classmethod
    def compatible_with_storage_def(cls, _):
        # Not needed for these tests
        raise NotImplementedError()

    @classmethod
    def set_object(cls, intermediate_store, obj, context, runtime_type, paths):
        check.inst_param(intermediate_store, 'intermediate_store', S3IntermediateStore)
        paths.append(obj)
        return intermediate_store.set_object('', context, runtime_type, paths)

    @classmethod
    def get_object(cls, intermediate_store, _context, _runtime_type, paths):
        check.inst_param(intermediate_store, 'intermediate_store', S3IntermediateStore)
        res = intermediate_store.object_store.s3.list_objects(
            Bucket=intermediate_store.object_store.bucket,
            Prefix=intermediate_store.key_for_paths(paths),
        )
        return res['Contents'][0]['Key'].split('/')[-1]


@aws
@nettest
def test_s3_intermediate_store_with_type_storage_plugin():
    run_id = str(uuid.uuid4())

    # FIXME need a dedicated test bucket
    intermediate_store = S3IntermediateStore(
        run_id=run_id,
        s3_bucket='dagster-airflow-scratch',
        type_storage_plugin_registry={RuntimeString.inst(): FancyStringS3TypeStoragePlugin},
    )

    with yield_empty_pipeline_context(run_id=run_id) as context:
        try:
            intermediate_store.set_value('hello', context, RuntimeString.inst(), ['obj_name'])

            assert intermediate_store.has_object(context, ['obj_name'])
            assert (
                intermediate_store.get_value(context, RuntimeString.inst(), ['obj_name']) == 'hello'
            )

        finally:
            intermediate_store.rm_object(context, ['obj_name'])


@aws
@nettest
def test_s3_intermediate_store_with_composite_type_storage_plugin():
    run_id = str(uuid.uuid4())

    # FIXME need a dedicated test bucket
    intermediate_store = S3IntermediateStore(
        run_id=run_id,
        s3_bucket='dagster-airflow-scratch',
        type_storage_plugin_registry={RuntimeString.inst(): FancyStringS3TypeStoragePlugin},
    )

    with yield_empty_pipeline_context(run_id=run_id) as context:
        with pytest.raises(check.NotImplementedCheckError):
            intermediate_store.set_value(
                ['hello'], context, resolve_to_runtime_type(List[String]), ['obj_name']
            )


@aws
@nettest
def test_s3_intermediate_store_composite_types_with_custom_serializer_for_inner_type():
    run_id = str(uuid.uuid4())

    intermediate_store = S3IntermediateStore(run_id=run_id, s3_bucket='dagster-airflow-scratch')
    with yield_empty_pipeline_context(run_id=run_id) as context:
        try:
            intermediate_store.set_object(
                ['foo', 'bar'],
                context,
                resolve_to_runtime_type(List[LowercaseString]).inst(),
                ['list'],
            )
            assert intermediate_store.has_object(context, ['list'])
            assert intermediate_store.get_object(
                context, resolve_to_runtime_type(List[Bool]).inst(), ['list']
            ) == ['foo', 'bar']

        finally:
            intermediate_store.rm_object(context, ['foo'])


@aws
@nettest
def test_s3_intermediate_store_with_custom_serializer():
    run_id = str(uuid.uuid4())

    # FIXME need a dedicated test bucket
    intermediate_store = S3IntermediateStore(run_id=run_id, s3_bucket='dagster-airflow-scratch')

    with yield_empty_pipeline_context(run_id=run_id) as context:
        try:
            intermediate_store.set_object('foo', context, LowercaseString.inst(), ['foo'])

            assert (
                intermediate_store.object_store.s3.get_object(
                    Bucket=intermediate_store.object_store.bucket,
                    Key='/'.join([intermediate_store.root] + ['foo']),
                )['Body']
                .read()
                .decode('utf-8')
                == 'FOO'
            )

            assert intermediate_store.has_object(context, ['foo'])
            assert intermediate_store.get_object(context, LowercaseString.inst(), ['foo']) == 'foo'
        finally:
            intermediate_store.rm_object(context, ['foo'])


@aws
@nettest
def test_s3_intermediate_store():
    run_id = str(uuid.uuid4())

    # FIXME need a dedicated test bucket
    intermediate_store = S3IntermediateStore(run_id=run_id, s3_bucket='dagster-airflow-scratch')
    assert intermediate_store.root == '/'.join(['dagster', 'runs', run_id, 'files'])

    with yield_empty_pipeline_context(run_id=run_id) as context:
        try:
            intermediate_store.set_object(True, context, RuntimeBool.inst(), ['true'])

            assert intermediate_store.has_object(context, ['true'])
            assert intermediate_store.get_object(context, RuntimeBool.inst(), ['true']) is True
            assert intermediate_store.uri_for_paths(['true']).startswith('s3://')

        finally:
            intermediate_store.rm_object(context, ['true'])



from dagster import pipeline, solid, ModeDefinition, execute_pipeline, ResourceDefinition
from dagster.utils.test import get_temp_file_handle_with_data
from dagster_aws.s3.solids import file_handle_to_s3
from dagster_aws.s3.s3_fake_resource import create_s3_fake_resource


def create_file_handle_pipeline(temp_file_handle, s3_resource):
    # pylint: disable=no-value-for-parameter

    @solid
    def emit_temp_handle(_):
        return temp_file_handle

    @pipeline(
        mode_defs=[
            ModeDefinition(resource_defs={'s3': ResourceDefinition.hardcoded_resource(s3_resource)})
        ]
    )
    def test():
        return file_handle_to_s3(emit_temp_handle())

    return test


def test_successful_file_handle_to_s3():
    foo_bytes = 'foo'.encode()
    with get_temp_file_handle_with_data(foo_bytes) as temp_file_handle:
        s3_fake_resource = create_s3_fake_resource()
        result = execute_pipeline(
            create_file_handle_pipeline(temp_file_handle, s3_fake_resource),
            environment_dict={
                'solids': {
                    'file_handle_to_s3': {'config': {'Bucket': 'some-bucket', 'Key': 'some-key'}}
                }
            },
        )

        assert result.success

        assert s3_fake_resource.session.mock_extras.put_object.call_count == 1

        assert (
            s3_fake_resource.session.get_object('some-bucket', 'some-key')['Body'].read()
            == foo_bytes
        )

        materializations = result.result_for_solid(
            'file_handle_to_s3'
        ).materializations_during_compute
        assert len(materializations) == 1
        assert len(materializations[0].metadata_entries) == 1
        assert (
            materializations[0].metadata_entries[0].entry_data.path == 's3://some-bucket/some-key'
        )
        assert materializations[0].metadata_entries[0].label == 'some-key'


def test_successful_file_handle_to_s3_with_configs():
    foo_bytes = 'foo'.encode()
    with get_temp_file_handle_with_data(foo_bytes) as temp_file_handle:
        s3_fake_resource = create_s3_fake_resource()

        result = execute_pipeline(
            create_file_handle_pipeline(temp_file_handle, s3_fake_resource),
            environment_dict={
                'solids': {
                    'file_handle_to_s3': {
                        'config': {
                            'Bucket': 'some-bucket',
                            'Key': 'some-key',
                            'CacheControl': 'some-value',
                        }
                    }
                }
            },
        )

        assert result.success

        s3_fake_resource.session.mock_extras.put_object.assert_called_once_with(
            CacheControl='some-value'
        )



import boto3

from botocore.handlers import disable_signing


def create_s3_session(signed=True):
    s3 = boto3.resource('s3').meta.client  # pylint:disable=C0103
    if not signed:
        s3.meta.events.register('choose-signer.s3.*', disable_signing)
    return s3


class S3Logger(object):
    def __init__(self, logger, bucket, key, filename, size):
        self._logger = logger
        self._bucket = bucket
        self._key = key
        self._filename = filename
        self._seen_so_far = 0
        self._size = size

    def __call__(self, bytes_amount):
        self._seen_so_far += bytes_amount
        percentage = (self._seen_so_far / self._size) * 100
        self._logger(
            'Download of {bucket}/{key} to {target_path}: {percentage}% complete'.format(
                bucket=self._bucket,
                key=self._key,
                target_path=self._filename,
                percentage=percentage,
            )
        )



from dagster import Field, resource

from .utils import create_s3_session


class S3Resource:
    def __init__(self, s3_session):
        self.session = s3_session

    def put_object(self, **kwargs):
        '''This mirrors the put_object boto3 API:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Bucket.put_object

        The config schema for this API is applied to the put_object_to_s3_bytes solid vs. at the
        resource level here.
        '''
        return self.session.put_object(**kwargs)


@resource(
    {
        'use_unsigned_session': Field(
            bool,
            description='Specifies whether to use an unsigned S3 session',
            is_optional=True,
            default_value=True,
        )
    }
)
def s3_resource(context):
    use_unsigned_session = context.resource_config['use_unsigned_session']
    return S3Resource(create_s3_session(signed=not use_unsigned_session))
